@@ -1,5 +1,6 @@
 /*
  * Copyright 2009 - 2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2011 by Alexey Osipov. <simba@lerlan.ru>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -21,18 +22,21 @@
 #endif
 
 #include "xf86Wacom.h"
+#include "wcmTouchFilter.h"
 #include <math.h>
 
 /* Defines for 2FC Gesture */
-#define WACOM_INLINE_DISTANCE        40
 #define WACOM_HORIZ_ALLOWED           1
 #define WACOM_VERT_ALLOWED            2
 #define WACOM_GESTURE_LAG_TIME       10
 
+#define GESTURE_NONE_MODE             0
 #define GESTURE_TAP_MODE              1
 #define GESTURE_SCROLL_MODE           2
 #define GESTURE_ZOOM_MODE             4
 #define GESTURE_LAG_MODE              8
+#define GESTURE_PREDRAG_MODE         16
+#define GESTURE_DRAG_MODE            32
 
 #define WCM_SCROLL_UP                 5	/* vertical up */
 #define WCM_SCROLL_DOWN               4	/* vertical down */
@@ -54,30 +58,37 @@ static Bool pointsInLine(WacomCommonPtr common, WacomDeviceState ds0,
 		WacomDeviceState ds1)
 {
 	Bool ret = FALSE;
+	Bool rotated = common->wcmRotate == ROTATE_CW ||
+			common->wcmRotate == ROTATE_CCW;
+	int horizon_rotated = (rotated) ?
+			WACOM_HORIZ_ALLOWED : WACOM_VERT_ALLOWED;
+	int vertical_rotated = (rotated) ?
+			WACOM_VERT_ALLOWED : WACOM_HORIZ_ALLOWED;
+	int max_spread = common->wcmGestureParameters.wcmMaxScrollFingerSpread;
 
 	if (!common->wcmGestureParameters.wcmScrollDirection)
 	{
-		if ((abs(ds0.x - ds1.x) < WACOM_INLINE_DISTANCE) &&
-			(abs(ds0.y - ds1.y) > WACOM_INLINE_DISTANCE))
+		if ((abs(ds0.x - ds1.x) < max_spread) &&
+			(abs(ds0.y - ds1.y) > max_spread))
 		{
-			common->wcmGestureParameters.wcmScrollDirection = WACOM_VERT_ALLOWED;
+			common->wcmGestureParameters.wcmScrollDirection = horizon_rotated;
 			ret = TRUE;
 		}
-		if ((abs(ds0.y - ds1.y) < WACOM_INLINE_DISTANCE) &&
-			(abs(ds0.x - ds1.x) > WACOM_INLINE_DISTANCE))
+		if ((abs(ds0.y - ds1.y) < max_spread) &&
+			(abs(ds0.x - ds1.x) > max_spread))
 		{
-			common->wcmGestureParameters.wcmScrollDirection = WACOM_HORIZ_ALLOWED;
+			common->wcmGestureParameters.wcmScrollDirection = vertical_rotated;
 			ret = TRUE;
 		}
 	}
-	else if (common->wcmGestureParameters.wcmScrollDirection == WACOM_HORIZ_ALLOWED)
+	else if (common->wcmGestureParameters.wcmScrollDirection == vertical_rotated)
 	{
-		if (abs(ds0.y - ds1.y) < WACOM_INLINE_DISTANCE)
+		if (abs(ds0.y - ds1.y) < max_spread)
 			ret = TRUE;
 	}
-	else if (common->wcmGestureParameters.wcmScrollDirection == WACOM_VERT_ALLOWED)
+	else if (common->wcmGestureParameters.wcmScrollDirection == horizon_rotated)
 	{
-		if (abs(ds0.x - ds1.x) < WACOM_INLINE_DISTANCE)
+		if (abs(ds0.x - ds1.x) < max_spread)
 			ret = TRUE;
 	}
 	return ret;
@@ -86,19 +97,11 @@ static Bool pointsInLine(WacomCommonPtr common, WacomDeviceState ds0,
 /* send a button event */
 static void wcmSendButtonClick(WacomDevicePtr priv, int button, int state)
 {
-	int x = 0;
-	int y = 0;
 	int mode = is_absolute(priv->pInfo);
 
-	if (mode)
-	{
-		x = priv->oldX;
-		y = priv->oldY;
-	}
-
 	/* send button event in state */
-	xf86PostButtonEvent(priv->pInfo->dev, mode,button,
-		state,0,priv->naxes,x,y,0,0,0,0);
+	xf86PostButtonEventP(priv->pInfo->dev, mode,button,
+		state,0,0,0);
 
 	/* We have changed the button state (from down to up) for the device
 	 * so we need to update the record */
@@ -140,34 +143,93 @@ static void wcmFingerTapToClick(WacomDevicePtr priv)
 	}
 }
 
-
-/* process single finger Relative mode events
- * if touch is not in an active gesture mode.
- */
-static void wcmFirstFingerClick(WacomCommonPtr common)
+static CARD32 wcmSingleFingerTapTimer(OsTimerPtr timer, CARD32 time, pointer arg)
 {
-	static int tmpStamp = 0;
-	WacomChannelPtr aChannel = common->wcmChannel;
-	WacomDeviceState ds = aChannel->valid.states[0];
-	WacomDeviceState dsLast = aChannel->valid.states[1];
-	if (ds.proximity)
+	WacomDevicePtr priv = (WacomDevicePtr)arg;
+	WacomCommonPtr common = priv->common;
+
+	if (common->wcmGestureMode == GESTURE_PREDRAG_MODE)
 	{
-		if (common->wcmTouchpadMode)
-			/* continuing left button down */
-			aChannel->valid.states[0].buttons |= 1;
-		else if (!dsLast.proximity &&
-		    (abs(tmpStamp - ds.sample) <= common->wcmGestureParameters.wcmTapTime))
-		{
-			/* initial left button down */
-			aChannel->valid.states[0].buttons |= 1;
-			common->wcmTouchpadMode = 1;
-		}
-	} else {
-		tmpStamp = GetTimeInMillis();
-		if (common->wcmTouchpadMode)
-			aChannel->valid.states[0].buttons &= ~1;
-		common->wcmTouchpadMode = 0;
+		/* left button down */
+		wcmSendButtonClick(priv, 1, 1);
+
+		/* left button up */
+		wcmSendButtonClick(priv, 1, 0);
+		common->wcmGestureMode = GESTURE_NONE_MODE;
 	}
+
+	return 0;
+}
+
+/* A single finger tap is defined as 1 finger tap that lasts less than
+ * wcmTapTime.  It results in a left button press.
+ *
+ * Some work must be done to make sure two fingers were not touching
+ * during this gesture. This is easy if first finger is released
+ * first.  To handle case of second finger released first, require
+ * second finger to have been released before first finger ever touched.
+ *
+ * Function relies on ds[0/1].sample to be updated only when entering or
+ * exiting proximity so no storage is needed when initial touch occurs.
+ */
+static void wcmSingleFingerTap(WacomDevicePtr priv)
+{
+	WacomCommonPtr common = priv->common;
+	WacomChannelPtr firstChannel = common->wcmChannel;
+	WacomChannelPtr secondChannel = common->wcmChannel + 1;
+	WacomDeviceState ds[2] = { firstChannel->valid.states[0],
+				   secondChannel->valid.states[0] };
+	WacomDeviceState dsLast[2] = { firstChannel->valid.states[1],
+					secondChannel->valid.states[1] };
+
+	DBG(10, priv, "\n");
+
+	/* This gesture is only valid on touchpads. */
+	if (TabletHasFeature(priv->common, WCM_LCD))
+		return;
+
+	if (!ds[0].proximity && dsLast[0].proximity && !ds[1].proximity)
+	{
+		/* Single Tap must have lasted less than wcmTapTime
+		 * and second finger must not have released after
+		 * first finger touched.
+		 */
+		if (ds[0].sample - dsLast[0].sample <=
+		    common->wcmGestureParameters.wcmTapTime &&
+		    ds[1].sample < dsLast[0].sample)
+		{
+			common->wcmGestureMode = GESTURE_PREDRAG_MODE;
+
+			/* Delay to detect possible drag operation */
+			TimerSet(NULL, 0, common->wcmGestureParameters.wcmTapTime, wcmSingleFingerTapTimer, priv);
+		}
+	}
+}
+
+/* Monitors for 1 finger touch and forces left button press or 1 finger
+ * release and will remove left button press.
+ *
+ * This function relies on wcmGestureMode will only be zero if
+ * WACOM_GESTURE_LAG_TIME has passed and still ony 1 finger on screen.
+ */
+static void wcmSingleFingerPress(WacomDevicePtr priv)
+{
+	WacomCommonPtr common = priv->common;
+	WacomChannelPtr firstChannel = common->wcmChannel;
+	WacomChannelPtr secondChannel = common->wcmChannel + 1;
+	WacomDeviceState ds[2] = { firstChannel->valid.states[0],
+				   secondChannel->valid.states[0] };
+
+	DBG(10, priv, "\n");
+
+	/* This gesture is only valid on touchscreens. */
+	if (!TabletHasFeature(priv->common, WCM_LCD))
+		return;
+
+	if (ds[0].proximity && !ds[1].proximity)
+		firstChannel->valid.states[0].buttons |= 1;
+	if (!ds[0].proximity && !ds[1].proximity)
+		firstChannel->valid.states[0].buttons &= ~1;
 }
 
 /* parsing gesture mode according to 2FGT data */
@@ -194,26 +256,47 @@ void wcmGestureFilter(WacomDevicePtr priv, int channel)
 	if (!common->wcmGesture)
 		goto ret;
 
-	/* second finger in prox. wait for gesture event if first finger
-	 * was in in prox */
-	if (ds[1].proximity && !common->wcmGestureMode && dsLast[0].proximity)
+	/* When 2 fingers are in proximity, it must always be in one of
+	 * the valid 2 fingers modes: LAG, SCROLL, or ZOOM.
+	 * LAG mode is used while deciding between SCROLL and ZOOM and
+	 * prevents cursor movement.  Force to LAG mode if ever in NONE
+	 * mode to stop cursor movement.
+	 */
+	if (ds[0].proximity && ds[1].proximity)
 	{
-		common->wcmTouchpadMode = 0;
-		common->wcmGestureMode = GESTURE_LAG_MODE;
-	}
-
-	/* first finger recently came in prox. But not the first time
-	 * wait for the second one for a certain time */
-	else if (dsLast[0].proximity &&
-	    ((GetTimeInMillis() - ds[0].sample) < WACOM_GESTURE_LAG_TIME))
-	{
-		if (!common->wcmGestureMode)
+		if (common->wcmGestureMode == GESTURE_NONE_MODE)
 			common->wcmGestureMode = GESTURE_LAG_MODE;
 	}
+	/* When only 1 finger is in proximity, it can be in either LAG mode,
+	 * NONE mode or DRAG mode.
+	 * 1 finger LAG mode is a very short time period mainly to debounce
+	 * initial touch.
+	 * NONE and DRAG mode means cursor is allowed to move around.
+	 * DRAG mode in addition means that left button pressed.
+	 * There is no need to bother about LAG_TIME while in DRAG mode.
+	 * TODO: This has to use dsLast[0] because of later logic that
+	 * wants mode to be NONE still when 1st entering proximity.
+	 * That could use some re-arranging/cleanup.
+	 *
+	 */
+	else if (dsLast[0].proximity && common->wcmGestureMode != GESTURE_DRAG_MODE)
+	{
+		CARD32 ms = GetTimeInMillis();
 
-	/* we've waited enough time */
-	else if (common->wcmGestureMode == GESTURE_LAG_MODE)
-		common->wcmGestureMode = 0;
+		if ((ms - ds[0].sample) < WACOM_GESTURE_LAG_TIME)
+		{
+			/* Must have recently come into proximity.  Change
+			 * into LAG mode.
+			 */
+			if (common->wcmGestureMode == GESTURE_NONE_MODE)
+				common->wcmGestureMode = GESTURE_LAG_MODE;
+		}
+		else
+		{
+			/* Been in LAG mode long enough. Force to NONE mode. */
+			common->wcmGestureMode = GESTURE_NONE_MODE;
+		}
+	}
 
 	if  (ds[1].proximity && !dsLast[1].proximity)
 	{
@@ -233,26 +316,40 @@ void wcmGestureFilter(WacomDevicePtr priv, int channel)
 		common->wcmGestureParameters.wcmGestureUsed  = 0;
 
 		/* initialize the cursor position */
-		if (!common->wcmGestureMode && !channel)
+		if (common->wcmGestureMode == GESTURE_NONE_MODE && !channel)
 			goto ret;
+
+		/* got second touch in TapTime interval after first one,
+		 * switch to DRAG mode */
+		if (common->wcmGestureMode == GESTURE_PREDRAG_MODE)
+		{
+			/* left button down */
+			wcmSendButtonClick(priv, 1, 1);
+			common->wcmGestureMode = GESTURE_DRAG_MODE;
+			goto ret;
+		}
 	}
 
 	if (!ds[0].proximity && !ds[1].proximity)
 	{
 		/* first finger was out-prox when GestureMode was still on */
-		if (!dsLast[0].proximity && common->wcmGestureMode)
+		if (!dsLast[0].proximity &&
+		    common->wcmGestureMode != GESTURE_NONE_MODE)
 			/* send first finger out prox */
 			wcmSoftOutEvent(priv->pInfo);
 
+		/* if were in DRAG mode, send left button up now */
+		if (common->wcmGestureMode == GESTURE_DRAG_MODE)
+			wcmSendButtonClick(priv, 1, 0);
+
 		/* exit gesture mode when both fingers are out */
-		common->wcmGestureMode = 0;
+		common->wcmGestureMode = GESTURE_NONE_MODE;
 		common->wcmGestureParameters.wcmScrollDirection = 0;
 
-		common->wcmTouchpadMode = 0;
 		goto ret;
 	}
 
-	if (!(common->wcmGestureMode & (GESTURE_SCROLL_MODE | GESTURE_ZOOM_MODE)))
+	if (!(common->wcmGestureMode & (GESTURE_SCROLL_MODE | GESTURE_ZOOM_MODE)) && channel)
 		wcmFingerTapToClick(priv);
 
 	/* Change mode happens only when both fingers are out */
@@ -273,22 +370,32 @@ void wcmGestureFilter(WacomDevicePtr priv, int channel)
 		    wcmFingerScroll(priv);
 
 	/* process complex two finger gestures */
-	else if ((2*common->wcmGestureParameters.wcmTapTime <
-	    (GetTimeInMillis() - ds[0].sample)) &&
-	    (2*common->wcmGestureParameters.wcmTapTime <
-	    (GetTimeInMillis() - ds[1].sample))
-	    && ds[0].proximity && ds[1].proximity)
-	{
-		/* scroll should be considered first since it requires
-		 * a finger distance check */
-		wcmFingerScroll(priv);
+	else {
+		CARD32 ms = GetTimeInMillis();
+		int taptime = common->wcmGestureParameters.wcmTapTime;
 
-		if (!(common->wcmGestureMode & GESTURE_SCROLL_MODE))
-		    wcmFingerZoom(priv);
+		if (ds[0].proximity && ds[1].proximity &&
+		    (taptime < (ms - ds[0].sample)) &&
+		    (taptime < (ms - ds[1].sample)))
+		{
+			/* scroll should be considered first since it requires
+			 * a finger distance check */
+			wcmFingerScroll(priv);
+
+			if (!(common->wcmGestureMode & GESTURE_SCROLL_MODE))
+				wcmFingerZoom(priv);
+		}
 	}
 ret:
-	if (!common->wcmGestureMode && !channel && !is_absolute(priv->pInfo))
-		wcmFirstFingerClick(common);
+	if (common->wcmGestureMode == GESTURE_NONE_MODE && !channel)
+	{
+		/* Since this is in ret block, can not rely on generic
+		 * wcmGesture enable check from above.
+		 */
+		if (common->wcmGesture)
+			wcmSingleFingerTap(priv);
+		wcmSingleFingerPress(priv);
+	}
 }
 
 static void wcmSendScrollEvent(WacomDevicePtr priv, int dist,
@@ -334,6 +441,7 @@ static void wcmFingerScroll(WacomDevicePtr priv)
 	int midPoint_old = 0;
 	int i = 0, dist = 0;
 	WacomFilterState filterd;  /* borrow this struct */
+	int max_spread = common->wcmGestureParameters.wcmMaxScrollFingerSpread;
 
 	DBG(10, priv, "\n");
 
@@ -341,7 +449,7 @@ static void wcmFingerScroll(WacomDevicePtr priv)
 	{
 		if (abs(touchDistance(ds[0], ds[1]) -
 			touchDistance(common->wcmGestureState[0],
-			common->wcmGestureState[1])) < WACOM_INLINE_DISTANCE)
+			common->wcmGestureState[1])) < max_spread)
 		{
 			/* two fingers stay close to each other all the time and
 			 * move in vertical or horizontal direction together
@@ -431,6 +539,7 @@ static void wcmFingerZoom(WacomDevicePtr priv)
 	int count, button;
 	int dist = touchDistance(common->wcmGestureState[0],
 			common->wcmGestureState[1]);
+	int max_spread = common->wcmGestureParameters.wcmMaxScrollFingerSpread;
 
 	DBG(10, priv, "\n");
 
@@ -440,13 +549,13 @@ static void wcmFingerZoom(WacomDevicePtr priv)
 		if (abs(touchDistance(ds[0], ds[1]) -
 			touchDistance(common->wcmGestureState[0],
 				      common->wcmGestureState[1])) >
-			(3 * WACOM_INLINE_DISTANCE))
+			(3 * max_spread))
 		{
 			/* left button might be down, send it up first */
 			wcmSendButtonClick(priv, 1, 0);
 
 			/* fingers moved apart more than 3 times
-			 * WACOM_INLINE_DISTANCE, zoom mode is entered */
+			 * wcmMaxScrollFingerSpread, zoom mode is entered */
 			common->wcmGestureMode = GESTURE_ZOOM_MODE;
 		}
 	}
@@ -487,4 +596,9 @@ static void wcmFingerZoom(WacomDevicePtr priv)
 	}
 }
 
+Bool wcmTouchNeedSendEvents(WacomCommonPtr common)
+{
+	return !(common->wcmGestureMode & ~GESTURE_DRAG_MODE);
+}
+
 /* vim: set noexpandtab tabstop=8 shiftwidth=8: */
