@@ -22,7 +22,6 @@
 #endif
 
 #include "xf86Wacom.h"
-#include "wcmFilter.h"
 
 #include <asm/types.h>
 #include <linux/input.h>
@@ -34,11 +33,16 @@
 typedef struct {
 	int wcmLastToolSerial;
 	int wcmBTNChannel;
+	int wcmDeviceType;
+	Bool wcmPenTouch;
 	Bool wcmUseMT;
 	int wcmMTChannel;
 	int wcmPrevChannel;
 	int wcmEventCnt;
 	struct input_event wcmEvents[MAX_USB_EVENTS];
+	int nbuttons;                /* total number of buttons */
+	int npadkeys;                /* number of pad keys in the above array */
+	int padkey_code[WCM_MAX_BUTTONS];/* hardware codes for buttons */
 } wcmUSBData;
 
 static Bool usbDetect(InputInfoPtr);
@@ -76,7 +80,6 @@ static struct _WacomModel mname =		\
 	.GetRanges = usbWcmGetRanges,		\
 	.Start = usbStart,			\
 	.Parse = usbParse,			\
-	.FilterRaw = wcmFilterCoord,		\
 	.DetectConfig = usbDetectConfig,	\
 };
 
@@ -95,6 +98,7 @@ DEFINE_MODEL(usbIntuos,		"USB Intuos1",		5);
 DEFINE_MODEL(usbIntuos2,	"USB Intuos2",		5);
 DEFINE_MODEL(usbIntuos3,	"USB Intuos3",		5);
 DEFINE_MODEL(usbIntuos4,	"USB Intuos4",		5);
+DEFINE_MODEL(usbIntuos5,	"USB Intuos5",		5);
 DEFINE_MODEL(usbVolito,		"USB Volito",		4);
 DEFINE_MODEL(usbVolito2,	"USB Volito2",		4);
 DEFINE_MODEL(usbCintiqV5,	"USB CintiqV5",		5);
@@ -200,7 +204,7 @@ static struct
 	{ WACOM_VENDOR_ID, 0xD7, 100000, 100000, &usbBamboo     }, /* CTH-461/S */
 	{ WACOM_VENDOR_ID, 0xD8, 100000, 100000, &usbBamboo     }, /* CTH-661/S1 */
 	{ WACOM_VENDOR_ID, 0xDA, 100000, 100000, &usbBamboo     }, /* CTH-461/L */
-	{ WACOM_VENDOR_ID, 0xDB, 100000, 100000, &usbBamboo     }, /* CTH-661 */
+	{ WACOM_VENDOR_ID, 0xDB, 100000, 100000, &usbBamboo     }, /* CTH-661/L */
 
 	{ WACOM_VENDOR_ID, 0x20, 100000, 100000, &usbIntuos     }, /* Intuos 4x5 */
 	{ WACOM_VENDOR_ID, 0x21, 100000, 100000, &usbIntuos     }, /* Intuos 6x8 */
@@ -219,6 +223,7 @@ static struct
 	{ WACOM_VENDOR_ID, 0x37,  20000,  20000, &usbCintiq     }, /* PL700 */
 	{ WACOM_VENDOR_ID, 0x38,  20000,  20000, &usbCintiq     }, /* PL510 */
 	{ WACOM_VENDOR_ID, 0x39,  20000,  20000, &usbCintiq     }, /* PL710 */
+	{ WACOM_VENDOR_ID, 0x3A,  20000,  20000, &usbCintiq     }, /* DTI520 */
 	{ WACOM_VENDOR_ID, 0xC0,  20000,  20000, &usbCintiq     }, /* DTF720 */
 	{ WACOM_VENDOR_ID, 0xC2,  20000,  20000, &usbCintiq     }, /* DTF720a */
 	{ WACOM_VENDOR_ID, 0xC4,  20000,  20000, &usbCintiq     }, /* DTF521 */
@@ -242,6 +247,8 @@ static struct
 
 	{ WACOM_VENDOR_ID, 0x65, 100000, 100000, &usbBamboo     }, /* Bamboo */
 	{ WACOM_VENDOR_ID, 0x69,  39842,  39842, &usbBamboo1    }, /* Bamboo1 */
+	{ WACOM_VENDOR_ID, 0x6A, 100000, 100000, &usbBamboo1    }, /* Bamboo1 4x6 */
+	{ WACOM_VENDOR_ID, 0x6B, 100000, 100000, &usbBamboo1    }, /* Bamboo1 5x8 */
 
 	{ WACOM_VENDOR_ID, 0xB0, 200000, 200000, &usbIntuos3    }, /* Intuos3 4x5 */
 	{ WACOM_VENDOR_ID, 0xB1, 200000, 200000, &usbIntuos3    }, /* Intuos3 6x8 */
@@ -258,17 +265,26 @@ static struct
 	{ WACOM_VENDOR_ID, 0xBC, 200000, 200000, &usbIntuos4    }, /* Intuos4 WL USB Endpoint */
 	{ WACOM_VENDOR_ID, 0xBD, 200000, 200000, &usbIntuos4    }, /* Intuos4 WL Bluetooth Endpoint */
 
+	{ WACOM_VENDOR_ID, 0x26, 200000, 200000, &usbIntuos5    }, /* Intuos5 touch S */
+	{ WACOM_VENDOR_ID, 0x27, 200000, 200000, &usbIntuos5    }, /* Intuos5 touch M */
+	{ WACOM_VENDOR_ID, 0x28, 200000, 200000, &usbIntuos5    }, /* Intuos5 touch L */
+	{ WACOM_VENDOR_ID, 0x29, 200000, 200000, &usbIntuos5    }, /* Intuos5 S */
+	{ WACOM_VENDOR_ID, 0x2A, 200000, 200000, &usbIntuos5    }, /* Intuos5 M */
+
 	{ WACOM_VENDOR_ID, 0x3F, 200000, 200000, &usbCintiqV5   }, /* Cintiq 21UX */
 	{ WACOM_VENDOR_ID, 0xC5, 200000, 200000, &usbCintiqV5   }, /* Cintiq 20WSX */
 	{ WACOM_VENDOR_ID, 0xC6, 200000, 200000, &usbCintiqV5   }, /* Cintiq 12WX */
 	{ WACOM_VENDOR_ID, 0xCC, 200000, 200000, &usbCintiqV5   }, /* Cintiq 21UX2 */
+	{ WACOM_VENDOR_ID, 0xF4, 200000, 200000, &usbCintiqV5   }, /* Cintiq 24HD */
 
 	{ WACOM_VENDOR_ID, 0x90, 100000, 100000, &usbTabletPC   }, /* TabletPC 0x90 */
 	{ WACOM_VENDOR_ID, 0x93, 100000, 100000, &usbTabletPC   }, /* TabletPC 0x93 */
+	{ WACOM_VENDOR_ID, 0x97, 100000, 100000, &usbTabletPC   }, /* TabletPC 0x97 */
 	{ WACOM_VENDOR_ID, 0x9A, 100000, 100000, &usbTabletPC   }, /* TabletPC 0x9A */
-	{ WACOM_VENDOR_ID, 0x9F,    394,    394, &usbTabletPC   }, /* CapPlus  0x9F */
-	{ WACOM_VENDOR_ID, 0xE2,    394,    394, &usbTabletPC   }, /* TabletPC 0xE2 */
+	{ WACOM_VENDOR_ID, 0x9F, 100000, 100000, &usbTabletPC   }, /* CapPlus  0x9F */
+	{ WACOM_VENDOR_ID, 0xE2, 100000, 100000, &usbTabletPC   }, /* TabletPC 0xE2 */
 	{ WACOM_VENDOR_ID, 0xE3, 100000, 100000, &usbTabletPC   }, /* TabletPC 0xE3 */
+	{ WACOM_VENDOR_ID, 0xE6, 100000, 100000, &usbTabletPC   }, /* TabletPC 0xE6 */
 
 	/* IDs from Waltop's driver, available http://www.waltop.com.tw/download.asp?lv=0&id=2.
 	   Accessed 8 Apr 2010, driver release date 2009/08/11, fork of linuxwacom 0.8.4.
@@ -315,6 +331,7 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
 	struct input_id sID;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
+	wcmUSBData *usbdata;
 
 	DBG(1, priv, "initializing USB tablet\n");
 
@@ -326,13 +343,14 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
 		return !Success;
 	}
 
+	usbdata = common->private;
 	*version = 0.0;
 
 	/* fetch vendor, product, and model name */
 	ioctl(pInfo->fd, EVIOCGID, &sID);
 	ioctl(pInfo->fd, EVIOCGNAME(sizeof(id)), id);
 
-	for (i = 0; i < sizeof (WacomModelDesc) / sizeof (WacomModelDesc [0]); i++)
+	for (i = 0; i < ARRAY_SIZE(WacomModelDesc); i++)
 	{
 		if (sID.vendor == WacomModelDesc[i].vendor_id &&
 		    sID.product == WacomModelDesc [i].model_id)
@@ -350,25 +368,25 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
 	}
 
 	/* Find out supported button codes. */
-	common->npadkeys = 0;
-	for (i = 0; i < sizeof (padkey_codes) / sizeof (padkey_codes [0]); i++)
+	usbdata->npadkeys = 0;
+	for (i = 0; i < ARRAY_SIZE(padkey_codes); i++)
 		if (ISBITSET (common->wcmKeys, padkey_codes [i]))
-			common->padkey_code [common->npadkeys++] = padkey_codes [i];
+			usbdata->padkey_code [usbdata->npadkeys++] = padkey_codes [i];
 
 	if (!(ISBITSET (common->wcmKeys, BTN_TOOL_MOUSE)))
 	{
 		/* If mouse buttons detected but no mouse tool
 		 * then they must be associated with pad buttons.
 		 */
-		for (i = sizeof(mouse_codes)/sizeof(mouse_codes[0]); i > 0; i--)
+		for (i = ARRAY_SIZE(mouse_codes) - 1; i > 0; i--)
 			if (ISBITSET(common->wcmKeys, mouse_codes[i]))
 				break;
 
 		/* Make sure room for fixed map mouse buttons.  This
 		 * means mappings may overlap with padkey_codes[].
 		 */
-		if (i != 0 && common->npadkeys < WCM_USB_MAX_MOUSE_BUTTONS)
-			common->npadkeys = WCM_USB_MAX_MOUSE_BUTTONS;
+		if (i != 0 && usbdata->npadkeys < WCM_USB_MAX_MOUSE_BUTTONS)
+			usbdata->npadkeys = WCM_USB_MAX_MOUSE_BUTTONS;
 	}
 
 	/* nbuttons tracks maximum buttons on all tools (stylus/mouse).
@@ -377,9 +395,9 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
 	 * Stylus support tip and 2 stlyus buttons.
 	 */
 	if (ISBITSET (common->wcmKeys, BTN_TOOL_MOUSE))
-		common->nbuttons = WCM_USB_MAX_MOUSE_BUTTONS;
+		usbdata->nbuttons = WCM_USB_MAX_MOUSE_BUTTONS;
 	else
-		common->nbuttons = WCM_USB_MAX_STYLUS_BUTTONS;
+		usbdata->nbuttons = WCM_USB_MAX_STYLUS_BUTTONS;
 
 	return Success;
 }
@@ -389,15 +407,10 @@ static void usbInitProtocol5(WacomCommonPtr common, const char* id,
 {
 	common->wcmProtocolLevel = WCM_PROTOCOL_5;
 	common->wcmPktLength = sizeof(struct input_event);
-	common->wcmCursorProxoutDistDefault 
-			= PROXOUT_INTUOS_DISTANCE;
+	common->wcmCursorProxoutDistDefault = PROXOUT_INTUOS_DISTANCE;
 
 	/* tilt enabled */
 	common->wcmFlags |= TILT_ENABLED_FLAG;
-
-	/* reinitialize max here since 0 is for Graphire series */
-	common->wcmMaxCursorDist = 256;
-
 }
 
 static void usbInitProtocol4(WacomCommonPtr common, const char* id,
@@ -405,13 +418,40 @@ static void usbInitProtocol4(WacomCommonPtr common, const char* id,
 {
 	common->wcmProtocolLevel = WCM_PROTOCOL_4;
 	common->wcmPktLength = sizeof(struct input_event);
-	common->wcmCursorProxoutDistDefault 
-			= PROXOUT_GRAPHIRE_DISTANCE;
+	common->wcmCursorProxoutDistDefault = PROXOUT_GRAPHIRE_DISTANCE;
 
 	/* tilt disabled */
 	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
+/* Initialize fixed PAD channel's state to in proximity.
+ *
+ * Some, but not all, Wacom protocol 4/5 devices are always in proximity.
+ * Because of evdev filtering, there will never be a BTN_TOOL_FINGER
+ * sent to initialize state.
+ * Generic protocol devices never send anything to help initialize PAD
+ * device as well.
+ * This helps those 2 cases and does not hurt the cases where kernel
+ * driver sends out-of-proximity event for PAD since PAD is always on
+ * its own channel, PAD_CHANNEL.
+ */
+static void usbWcmInitPadState(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
+	WacomCommonPtr common = priv->common;
+	WacomDeviceState *ds;
+	int channel = PAD_CHANNEL;
+
+	DBG(6, common, "Initializing PAD channel %d\n", channel);
+
+	ds = &common->wcmChannel[channel].work;
+
+	ds->proximity = 1;
+	ds->device_type = PAD_ID;
+	ds->device_id = PAD_DEVICE_ID;
+	ds->serial_num = channel;
+}
+
 int usbWcmGetRanges(InputInfoPtr pInfo)
 {
 	struct input_absinfo absinfo;
@@ -436,14 +476,14 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 		return !Success;
 	}
 
-        if (ioctl(pInfo->fd, EVIOCGBIT(EV_ABS,sizeof(abs)),abs) < 0)
+	if (!ISBITSET(ev,EV_ABS))
 	{
-		xf86Msg(X_ERROR, "%s: unable to ioctl abs bits.\n", pInfo->name);
+		xf86Msg(X_ERROR, "%s: no abs bits.\n", pInfo->name);
 		return !Success;
 	}
 
 	/* absolute values */
-	if (!ISBITSET(ev,EV_ABS))
+        if (ioctl(pInfo->fd, EVIOCGBIT(EV_ABS, sizeof(abs)), abs) < 0)
 	{
 		xf86Msg(X_ERROR, "%s: unable to ioctl max values.\n", pInfo->name);
 		return !Success;
@@ -458,17 +498,27 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (absinfo.maximum <= 0)
 	{
-		xf86Msg(X_ERROR, "%s: xmax value is wrong.\n", pInfo->name);
+		xf86Msg(X_ERROR, "%s: xmax value is %d, expected > 0.\n",
+			pInfo->name, absinfo.maximum);
 		return !Success;
 	}
+
 	if (!is_touch)
+	{
 		common->wcmMaxX = absinfo.maximum;
+
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
+		if (absinfo.resolution > 0)
+			common->wcmResolX = absinfo.resolution * 1000;
+#endif
+	}
 	else
 	{
 		common->wcmMaxTouchX = absinfo.maximum;
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
-		common->wcmTouchResolX = absinfo.resolution * 1000;
+		if (absinfo.resolution > 0)
+			common->wcmTouchResolX = absinfo.resolution * 1000;
 #endif
 	}
 
@@ -481,23 +531,34 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (absinfo.maximum <= 0)
 	{
-		xf86Msg(X_ERROR, "%s: ymax value is wrong.\n", pInfo->name);
+		xf86Msg(X_ERROR, "%s: ymax value is %d, expected > 0.\n",
+			pInfo->name, absinfo.maximum);
 		return !Success;
 	}
+
 	if (!is_touch)
+	{
 		common->wcmMaxY = absinfo.maximum;
+
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
+		if (absinfo.resolution > 0)
+			common->wcmResolY = absinfo.resolution * 1000;
+#endif
+	}
 	else
 	{
 		common->wcmMaxTouchY = absinfo.maximum;
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
-		common->wcmTouchResolY = absinfo.resolution * 1000;
+		if (absinfo.resolution > 0)
+			common->wcmTouchResolY = absinfo.resolution * 1000;
 #endif
 	}
 
 	/* max finger strip X for tablets with Expresskeys
 	 * or physical X for touch devices in hundredths of a mm */
-	if (ioctl(pInfo->fd, EVIOCGABS(ABS_RX), &absinfo) == 0)
+	if (ISBITSET(abs, ABS_RX) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_RX), &absinfo))
 	{
 		if (is_touch)
 			common->wcmTouchResolX =
@@ -509,7 +570,8 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	/* max finger strip Y for tablets with Expresskeys
 	 * or physical Y for touch devices in hundredths of a mm */
-	if (ioctl(pInfo->fd, EVIOCGABS(ABS_RY), &absinfo) == 0)
+	if (ISBITSET(abs, ABS_RY) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_RY), &absinfo))
 	{
 		if (is_touch)
 			common->wcmTouchResolY =
@@ -520,20 +582,30 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 	}
 
 	/* max z cannot be configured */
-	if (ioctl(pInfo->fd, EVIOCGABS(ABS_PRESSURE), &absinfo) == 0)
+	if (ISBITSET(abs, ABS_PRESSURE) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_PRESSURE), &absinfo))
 		common->wcmMaxZ = absinfo.maximum;
 
 	/* max distance */
-	if (ioctl(pInfo->fd, EVIOCGABS(ABS_DISTANCE), &absinfo) == 0)
+	if (ISBITSET(abs, ABS_DISTANCE) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_DISTANCE), &absinfo))
 		common->wcmMaxDist = absinfo.maximum;
 
 	if (ISBITSET(abs, ABS_MT_SLOT))
+	{
 		private->wcmUseMT = 1;
 
+		/* pen and MT on the same logical port */
+		if (ISBITSET(common->wcmKeys, BTN_TOOL_PEN))
+			private->wcmPenTouch = TRUE;
+	}
+
 	/* A generic protocol device does not report ABS_MISC event */
 	if (!ISBITSET(abs, ABS_MISC))
 		common->wcmProtocolLevel = WCM_PROTOCOL_GENERIC;
 
+	usbWcmInitPadState(pInfo);
+
 	return Success;
 }
 
@@ -541,12 +613,13 @@ static int usbDetectConfig(InputInfoPtr pInfo)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
+	wcmUSBData *usbdata = common->private;
 
 	DBG(10, common, "\n");
 	if (IsPad (priv))
-		priv->nbuttons = common->npadkeys;
+		priv->nbuttons = usbdata->npadkeys;
 	else
-		priv->nbuttons = common->nbuttons;
+		priv->nbuttons = usbdata->nbuttons;
 
 	if (!common->wcmCursorProxoutDist)
 		common->wcmCursorProxoutDist
@@ -574,7 +647,7 @@ static int usbChooseChannel(WacomCommonPtr common)
 	/* figure out the channel to use based on serial number */
 	int i, channel = -1;
 	wcmUSBData* private = common->private;
-	int serial = private->wcmLastToolSerial;
+	unsigned int serial = private->wcmLastToolSerial;
 
 	if (common->wcmProtocolLevel == WCM_PROTOCOL_GENERIC)
 	{
@@ -820,12 +893,29 @@ static int usbFilterEvent(WacomCommonPtr common, struct input_event *event)
 		}
 		else if (event->type == EV_ABS)
 		{
-			switch(event->code)
+			if (private->wcmDeviceType == TOUCH_ID)
 			{
-				case ABS_X:
-				case ABS_Y:
-				case ABS_PRESSURE:
-					return 1;
+				/* filter ST for MT */
+				switch(event->code)
+				{
+					case ABS_X:
+					case ABS_Y:
+					case ABS_PRESSURE:
+						return 1;
+				}
+			}
+			else
+			{
+				/* filter MT for pen */
+				switch(event->code)
+				{
+					case ABS_MT_SLOT:
+					case ABS_MT_TRACKING_ID:
+					case ABS_MT_POSITION_X:
+					case ABS_MT_POSITION_Y:
+					case ABS_MT_PRESSURE:
+						return 1;
+				}
 			}
 		}
 	}
@@ -849,6 +939,84 @@ static int usbFilterEvent(WacomCommonPtr common, struct input_event *event)
 	return 0;
 }
 
+#define ERASER_BIT      0x008
+#define PUCK_BITS	0xf00
+#define PUCK_EXCEPTION  0x806
+/**
+ * Decide the tool type by its id for protocol 5 devices
+ *
+ * @param id The tool id received from the kernel.
+ * @return The tool type associated with the tool id.
+ */
+static int usbIdToType(int id)
+{
+	int type = STYLUS_ID;
+
+	/* The existing tool ids have the following patten: all pucks, except
+	 * one, have the third byte set to zero; all erasers have the fourth
+	 * bit set. The rest are styli.
+	 */
+	if (id & ERASER_BIT)
+		type = ERASER_ID;
+	else if (!(id & PUCK_BITS) || (id == PUCK_EXCEPTION))
+		type = CURSOR_ID;
+
+	return type;
+}
+
+/**
+ * Find the tool type (STYLUS_ID, etc.) based on the device_id or the
+ *  current tool serial number if the device_id is unknown (0).
+ *
+ * Protocol 5 devices report different IDs for different styli and pucks,
+ * Protocol 4 devices simply report STYLUS_DEVICE_ID, etc.
+ *
+ * @param ds The current device state received from the kernel.
+ * @return The tool type associated with the tool id or the current
+ * tool serial number.
+ */
+static int usbFindDeviceType(const WacomCommonPtr common,
+			  const WacomDeviceState *ds)
+{
+	WacomToolPtr tool = NULL;
+	int device_type = 0;
+
+	if (!ds->device_id && ds->serial_num)
+	{
+		for (tool = common->wcmTool; tool; tool = tool->next)
+			if (ds->serial_num == tool->serial)
+			{
+				device_type = tool->typeid;
+				break;
+			}
+	}
+
+	if (device_type || !ds->device_id) return device_type;
+
+	switch (ds->device_id)
+	{
+		case STYLUS_DEVICE_ID:
+			device_type = STYLUS_ID;
+			break;
+		case ERASER_DEVICE_ID:
+			device_type = ERASER_ID;
+			break;
+		case CURSOR_DEVICE_ID:
+			device_type = CURSOR_ID;
+			break;
+		case TOUCH_DEVICE_ID:
+			device_type = TOUCH_ID;
+			break;
+		case PAD_DEVICE_ID:
+			device_type = PAD_ID;
+			break;
+		default: /* protocol 5 */
+			device_type = usbIdToType(ds->device_id);
+	}
+
+	return device_type;
+}
+
 static int usbParseAbsEvent(WacomCommonPtr common,
 			    struct input_event *event, WacomDeviceState *ds)
 {
@@ -878,10 +1046,7 @@ static int usbParseAbsEvent(WacomCommonPtr common,
 			ds->tilty = event->value - common->wcmMaxtiltY/2;
 			break;
 		case ABS_PRESSURE:
-			if (ds->device_type == TOUCH_ID)
-				ds->capacity = event->value;
-			else
-				ds->pressure = event->value;
+			ds->pressure = event->value;
 			break;
 		case ABS_DISTANCE:
 			ds->distance = event->value;
@@ -893,11 +1058,19 @@ static int usbParseAbsEvent(WacomCommonPtr common,
 			ds->abswheel = event->value;
 			break;
 		case ABS_THROTTLE:
-			ds->throttle = event->value;
+			/* 2nd touch ring comes in over ABS_THROTTLE for 24HD */
+			if (common->vendor_id == WACOM_VENDOR_ID && common->tablet_id == 0xF4)
+				ds->abswheel2 = event->value;
+			else
+				ds->throttle = event->value;
 			break;
 		case ABS_MISC:
+			ds->proximity = (event->value != 0);
 			if (event->value)
+			{
 				ds->device_id = event->value;
+				ds->device_type = usbFindDeviceType(common, ds);
+			}
 			break;
 		default:
 			change = 0;
@@ -905,6 +1078,36 @@ static int usbParseAbsEvent(WacomCommonPtr common,
 	return change;
 }
 
+/**
+ * Flip the mask bit in buttons corresponding to btn to the specified state.
+ *
+ * @param buttons The current button mask
+ * @param btn Zero-indexed button number to change
+ * @param state Zero to unset, non-zero to set the mask for the button
+ *
+ * @return The new button mask
+ */
+static int mod_buttons(int buttons, int btn, int state)
+{
+	int mask;
+
+	if (btn >= sizeof(int) * 8)
+	{
+		xf86Msg(X_ERROR, "%s: Invalid button number %d. Insufficient "
+				"storage\n", __func__, btn);
+		return buttons;
+	}
+
+	mask = 1 << btn;
+
+	if (state)
+		buttons |= mask;
+	else
+		buttons &= ~mask;
+
+	return buttons;
+}
+
 static int usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 {
 	int change = 1;
@@ -937,7 +1140,7 @@ static int usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 			break;
 
 		case ABS_MT_PRESSURE:
-			ds->capacity = event->value;
+			ds->pressure = event->value;
 			break;
 
 		default:
@@ -965,17 +1168,10 @@ static struct
 	{ PAD_ID,    BTN_0              }
 };
 
-#define MOD_BUTTONS(bit, value) do { \
-	shift = 1<<bit; \
-	ds->buttons = (((value) != 0) ? \
-		       (ds->buttons | (shift)) : (ds->buttons & ~(shift))); \
-        } while (0)
-
 static int usbParseKeyEvent(WacomCommonPtr common,
 			    struct input_event *event, WacomDeviceState *ds,
 			    WacomDeviceState *dslast)
 {
-	int shift;
 	int change = 1;
 
 	/* BTN_TOOL_* are sent to indicate when a specific tool is going
@@ -1038,7 +1234,6 @@ static int usbParseKeyEvent(WacomCommonPtr common,
 					ds->device_type = TOUCH_ID;
 					ds->device_id = TOUCH_DEVICE_ID;
 					ds->proximity = event->value;
-					MOD_BUTTONS(0, event->value);
 				}
 			}
 			break;
@@ -1068,18 +1263,6 @@ static int usbParseKeyEvent(WacomCommonPtr common,
 			if ((ds->proximity && !dslast->proximity) ||
 			    (!ds->proximity && dslast->proximity))
 				ds->sample = (int)GetTimeInMillis();
-			/* left button is always pressed for
-			 * touchscreen without capacity
-			 * when the first finger touch event received.
-			 * For touchscreen with capacity, left button
-			 * event will be decided
-			 * in wcmCommon.c by capacity threshold.
-			 * Touchpads should not have button
-			 * press.
-			 */
-			if (common->wcmCapacityDefault < 0 &&
-			    (TabletHasFeature(common, WCM_LCD)))
-				MOD_BUTTONS(0, event->value);
 			break;
 
 		case BTN_TOOL_TRIPLETAP:
@@ -1114,11 +1297,11 @@ static int usbParseKeyEvent(WacomCommonPtr common,
 	switch (event->code)
 	{
 		case BTN_STYLUS:
-			MOD_BUTTONS(1, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 1, event->value);
 			break;
 
 		case BTN_STYLUS2:
-			MOD_BUTTONS(2, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 2, event->value);
 			break;
 
 		default:
@@ -1132,48 +1315,117 @@ static int usbParseKeyEvent(WacomCommonPtr common,
 static int usbParseBTNEvent(WacomCommonPtr common,
 			    struct input_event *event, WacomDeviceState *ds)
 {
-	int shift, nkeys;
+	int nkeys;
 	int change = 1;
+	wcmUSBData *usbdata = common->private;
 
 	switch (event->code)
 	{
 		case BTN_LEFT:
-			MOD_BUTTONS(0, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 0, event->value);
 			break;
 
 		case BTN_MIDDLE:
-			MOD_BUTTONS(1, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 1, event->value);
 			break;
 
 		case BTN_RIGHT:
-			MOD_BUTTONS(2, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 2, event->value);
 			break;
 
 		case BTN_SIDE:
 		case BTN_BACK:
-			MOD_BUTTONS(3, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 3, event->value);
 			break;
 
 		case BTN_EXTRA:
 		case BTN_FORWARD:
-			MOD_BUTTONS(4, event->value);
+			ds->buttons = mod_buttons(ds->buttons, 4, event->value);
 			break;
 
 		default:
-			for (nkeys = 0; nkeys < common->npadkeys; nkeys++)
+			for (nkeys = 0; nkeys < usbdata->npadkeys; nkeys++)
 			{
-				if (event->code == common->padkey_code[nkeys])
+				if (event->code == usbdata->padkey_code[nkeys])
 				{
-					MOD_BUTTONS(nkeys, event->value);
+					ds->buttons = mod_buttons(ds->buttons, nkeys, event->value);
 					break;
 				}
 			}
-			if (nkeys >= common->npadkeys)
+			if (nkeys >= usbdata->npadkeys)
 				change = 0;
 	}
 	return change;
 }
 
+/***
+ * Retrieve the tool type from an USB data packet by looking at the event
+ * codes. Refer to linux/input.h for event codes that define tool types.
+ *
+ * @param event_ptr A pointer to the USB data packet that contains the
+ * events to be processed.
+ * @param nevents Number of events in the packet.
+ * @param last_device_type The device type for the last event
+ *
+ * @return The tool type. last_device_type if no pen/touch/eraser event code
+ *         in the event, or TOUCH_ID if last_device_type is not a tool.
+ */
+static int usbInitToolType(const struct input_event *event_ptr, int nevents, int last_device_type)
+{
+	int i, device_type = 0;
+	struct input_event* event = (struct input_event *)event_ptr;
+
+	for (i = 0; (i < nevents) && !device_type; ++i)
+	{
+		switch (event->code)
+		{
+			case BTN_TOOL_PEN:
+			case BTN_TOOL_PENCIL:
+			case BTN_TOOL_BRUSH:
+			case BTN_TOOL_AIRBRUSH:
+				device_type = STYLUS_ID;
+				break;
+
+			case BTN_TOOL_FINGER:
+			case ABS_MT_SLOT:
+			case ABS_MT_TRACKING_ID:
+				device_type = TOUCH_ID;
+				break;
+
+			case BTN_TOOL_RUBBER:
+				device_type = ERASER_ID;
+				break;
+		}
+
+		event++;
+	}
+
+	if (!device_type)
+	{
+		if (last_device_type)
+			device_type = last_device_type;
+		else
+			device_type = TOUCH_ID;
+	}
+
+	return device_type;
+}
+
+/**
+ * Check if the tool is a stylus/eraser and in-prox or not.
+ *
+ * @param device_type The tool type stored in wcmChannel
+ * @param proximity The tool's proximity state
+
+ * @return True if stylus/eraser is in-prox; False otherwise.
+ */
+static Bool usbIsPenInProx(int device_type, int proximity)
+{
+	Bool is_pen = (device_type == STYLUS_ID) ||
+			(device_type == ERASER_ID);
+	return (is_pen && proximity);
+}
+
 static void usbDispatchEvents(InputInfoPtr pInfo)
 {
 	int i;
@@ -1183,11 +1435,33 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	WacomCommonPtr common = priv->common;
 	int channel;
 	int channel_change = 0, btn_channel_change = 0, mt_channel_change = 0;
-	WacomDeviceState dslast;
+	WacomDeviceState dslast = common->wcmChannel[0].valid.state;
 	wcmUSBData* private = common->private;
 
 	DBG(6, common, "%d events received\n", private->wcmEventCnt);
 
+	if (private->wcmUseMT)
+		private->wcmDeviceType = usbInitToolType(private->wcmEvents,
+							 private->wcmEventCnt,
+							 dslast.device_type);
+
+	if (private->wcmPenTouch)
+	{
+		/* We get both pen and touch data from the kernel when they
+		 * both are in/down. So, if we were (hence the need of dslast)
+		 * processing pen events, we should ignore touch events.
+		 *
+		 * MT events will be posted to the userland when XInput 2.1
+		 * is ready.
+		 */
+		if ((private->wcmDeviceType == TOUCH_ID) &&
+				usbIsPenInProx(dslast.device_type, dslast.proximity))
+		{
+			private->wcmEventCnt = 0;
+			return;
+		}
+	}
+
 	channel = usbChooseChannel(common);
 
 	/* couldn't decide channel? invalid data */
@@ -1297,11 +1571,18 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	if (!ds->device_type && !dslast.proximity)
 	{
 		unsigned long keys[NBITS(KEY_MAX)] = { 0 };
+		int rc;
 
 		/* Retrieve the type by asking a resend from the kernel */
-		ioctl(common->fd, EVIOCGKEY(sizeof(keys)), keys);
+		rc = ioctl(common->fd, EVIOCGKEY(sizeof(keys)), keys);
+		if (rc == -1)
+		{
+			xf86Msg(X_ERROR, "%s: failed to retrieve key bits\n",
+					pInfo->name);
+			return;
+		}
 
-		for (i=0; i<sizeof(wcmTypeToKey) / sizeof(wcmTypeToKey[0]); i++)
+		for (i = 0; i < ARRAY_SIZE(wcmTypeToKey); i++)
 		{
 			if (ISBITSET(keys, wcmTypeToKey[i].tool_key))
 			{
@@ -1312,10 +1593,6 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 		}
 	}
 
-	/* don't send touch event when touch isn't enabled */
-	if ((ds->device_type == TOUCH_ID) && !common->wcmTouch)
-		return;
-
 	/* DTF720 and DTF720a don't support eraser */
 	if (((common->tablet_id == 0xC0) || (common->tablet_id == 0xC2)) && 
 		(ds->device_type == ERASER_ID)) 
@@ -1329,34 +1606,43 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	if (!ds->proximity)
 		private->wcmLastToolSerial = 0;
 
-	/* dispatch events */
-	if (channel_change ||
-	    (private->wcmBTNChannel == channel && btn_channel_change))
-		wcmEvent(common, channel, ds);
-
-	/* dispatch for second finger.  first finger is handled above. */
-	if (mt_channel_change)
+	/* don't send touch event when touch isn't enabled */
+	if (ds->device_type != TOUCH_ID || common->wcmTouch)
 	{
-		WacomDeviceState *mt_ds;
+		/* dispatch events */
+		if (channel_change ||
+		    (private->wcmBTNChannel == channel && btn_channel_change))
+			wcmEvent(common, channel, ds);
 
-		mt_ds = &common->wcmChannel[1].work;
-		wcmEvent(common, 1, mt_ds);
+		/* dispatch for second finger.
+		 * first finger is handled above. */
+		if (mt_channel_change)
+		{
+			WacomDeviceState *mt_ds;
+
+			mt_ds = &common->wcmChannel[1].work;
+			wcmEvent(common, 1, mt_ds);
+		}
 	}
 
        /* dispatch butten events when re-routed */
 	if (private->wcmBTNChannel != channel && btn_channel_change)
-	{
-		/* Force to in proximity for this special case */
-		btn_ds->proximity = 1;
-		btn_ds->device_type = PAD_ID;
-		btn_ds->device_id = PAD_DEVICE_ID;
-		btn_ds->serial_num = 0xf0;
 		wcmEvent(common, private->wcmBTNChannel, btn_ds);
-	}
 }
 
-/* Quirks to unify the tool types for GENERIC protocol tablet PCs */
-static void usbGenericTouchscreenQuirks(unsigned long *keys, unsigned long *abs)
+/* Quirks to unify the tool and tablet types for GENERIC protocol tablet PCs
+ *
+ * @param[in,out] keys Contains keys queried from hardware. If a
+ *   touchscreen is detected, keys are modified to add BTN_TOOL_FINGER so
+ *   that a TOUCH device is created later.
+ * @param[in] abs Used to detect multi-touch touchscreens.  When detected,
+ *   updates keys to add possibly missing BTN_TOOL_DOUBLETAP.
+ * @param[in,out] common Used only for tablet features.  Adds TCM_TPC for
+ *   touchscreens so correct defaults, such as absolute mode, are used.
+ */
+static void usbGenericTouchscreenQuirks(unsigned long *keys,
+					unsigned long *abs,
+					WacomCommonPtr common)
 {
 	/* USB Tablet PC single finger touch devices do not emit
 	 * BTN_TOOL_FINGER since it is a touchscreen device.
@@ -1364,7 +1650,10 @@ static void usbGenericTouchscreenQuirks(unsigned long *keys, unsigned long *abs)
 	if (ISBITSET(keys, BTN_TOUCH) &&
 			!ISBITSET(keys, BTN_TOOL_FINGER) &&
 			!ISBITSET(keys, BTN_TOOL_PEN))
+	{
 		SETBIT(keys, BTN_TOOL_FINGER); /* 1FGT */
+		TabletSetFeature(common, WCM_TPC);
+	}
 
 	/* Serial Tablet PC two finger touch devices do not emit
 	 * BTN_TOOL_DOUBLETAP since they are not touchpads.
@@ -1414,9 +1703,12 @@ static int usbProbeKeys(InputInfoPtr pInfo)
 	if (!ISBITSET(abs, ABS_MISC))
 	{
 		common->wcmProtocolLevel = WCM_PROTOCOL_GENERIC;
-		usbGenericTouchscreenQuirks(common->wcmKeys, abs);
+		usbGenericTouchscreenQuirks(common->wcmKeys, abs, common);
 	}
 
+	common->vendor_id = wacom_id.vendor;
+	common->tablet_id = wacom_id.product;
+
 	return wacom_id.product;
 }
 
