@@ -23,6 +23,7 @@
 /*****************************************************************************
  * General Defines
  ****************************************************************************/
+#include <wacom-util.h>
 #include <asm/types.h>
 #include <linux/input.h>
 #define MAX_USB_EVENTS 32
@@ -45,6 +46,9 @@
 #define MIN_PAD_RING 0		/* I4 absolute scroll ring min value */
 #define MAX_PAD_RING 71		/* I4 absolute scroll ring max value */
 
+/* I4 cursor tool has a rotation offset of 175 degrees */
+#define INTUOS4_CURSOR_ROTATION_OFFSET 175
+
 /* Default max distance to the tablet at which a proximity-out event is generated for
  * cursor device (e.g. mouse). 
  */
@@ -91,16 +95,6 @@
 #define ERASER_PROX     4
 #define OTHER_PROX      1
 
-/* to access kernel defined bits */
-#define BIT(x)		(1UL<<((x) & (BITS_PER_LONG - 1)))
-#define BITS_PER_LONG	(sizeof(long) * 8)
-#define NBITS(x)	((((x)-1)/BITS_PER_LONG)+1)
-#define ISBITSET(x,y)	((x)[LONG(y)] & BIT(y))
-#define SETBIT(x,y)	((x)[LONG(y)] |= BIT(y))
-#define CLEARBIT(x,y)	((x)[LONG(y)] &= ~BIT(y))
-#define OFF(x)		((x)%BITS_PER_LONG)
-#define LONG(x)		((x)/BITS_PER_LONG)
-
 /******************************************************************************
  * Forward Declarations
  *****************************************************************************/
@@ -128,8 +122,6 @@ struct _WacomModel
 	int (*GetRanges)(InputInfoPtr pInfo);
 	int (*Start)(InputInfoPtr pInfo);
 	int (*Parse)(InputInfoPtr pInfo, const unsigned char* data, int len);
-	int (*FilterRaw)(WacomCommonPtr common, WacomChannelPtr pChannel,
-		WacomDeviceStatePtr ds);
 	int (*DetectConfig)(InputInfoPtr pInfo);
 };
 
@@ -137,6 +129,7 @@ struct _WacomModel
  * WacomDeviceRec
  *****************************************************************************/
 
+/* these device IDs are reported through ABS_MISC for Protocol 4 devices */
 #define DEVICE_ID(flags) ((flags) & 0xff)
 #define STYLUS_DEVICE_ID	0x02
 #define TOUCH_DEVICE_ID		0x03
@@ -171,10 +164,12 @@ struct _WacomModel
 #define WCM_TPC			(0x00000200 | WCM_LCD) /* TabletPC (special
 							  button handling,
 							  always an LCD) */
-#define TabletHasFeature(common, feature) (((common)->tablet_type & (feature)) != 0)
+#define WCM_PENTOUCH		0x00000400 /* Tablet supports pen and touch */
+#define WCM_DUALRING		0x00000800 /* Tablet has two touch rings */
+#define TabletHasFeature(common, feature) MaskIsSet((common)->tablet_type, (feature))
+#define TabletSetFeature(common, feature) MaskSet((common)->tablet_type, (feature))
 
 #define ABSOLUTE_FLAG		0x00000100
-#define KEEP_SHAPE_FLAG		0x00000200
 #define BAUD_19200_FLAG		0x00000400
 #define BUTTONS_ONLY_FLAG	0x00000800
 
@@ -183,17 +178,20 @@ struct _WacomModel
 #define IsTouch(priv)  (DEVICE_ID((priv)->flags) == TOUCH_ID)
 #define IsEraser(priv) (DEVICE_ID((priv)->flags) == ERASER_ID)
 #define IsPad(priv)    (DEVICE_ID((priv)->flags) == PAD_ID)
+#define IsPen(priv)    (IsStylus(priv) || IsEraser(priv))
 
 #define IsUSBDevice(common) ((common)->wcmDevCls == &gWacomUSBDevice)
 
 #define FILTER_PRESSURE_RES	2048	/* maximum points in pressure curve */
+/* Tested result for setting the pressure threshold to a reasonable value */
+#define THRESHOLD_TOLERANCE (FILTER_PRESSURE_RES / 125)
+#define DEFAULT_THRESHOLD (FILTER_PRESSURE_RES / 75)
+
 #define WCM_MAX_BUTTONS		32	/* maximum number of tablet buttons */
-#define WCM_MAX_MOUSE_BUTTONS	16	/* maximum number of buttons-on-pointer
-                                         * (which are treated as mouse buttons,
-                                         * not as keys like tablet menu buttons). 
-					 * For backword compability support, 
-					 * tablet buttons besides the strips are
-					 * treated as buttons */
+
+#define AXIS_INVERT  0x01               /* Flag describing an axis which increases "downward" */
+#define AXIS_BITWISE 0x02               /* Flag describing an axis which changes bitwise */
+
 /* get/set/property */
 typedef struct _PROPINFO PROPINFO;
 
@@ -209,7 +207,7 @@ struct _PROPINFO
 
 struct _WacomDeviceRec
 {
-	char *name;		/* Do not move, same offset as common->wcmDevice. Used by DBG macro */
+	char *name;		/* Do not move, same offset as common->device_path. Used by DBG macro */
 	/* configuration fields */
 	struct _WacomDeviceRec *next;
 	InputInfoPtr pInfo;
@@ -226,7 +224,9 @@ struct _WacomDeviceRec
 	int maxY;	        /* tool physical maxY in device coordinates*/
 	double factorX;		/* X factor */
 	double factorY;		/* Y factor */
-	unsigned int serial;	/* device serial number */
+	unsigned int serial;	/* device serial number this device takes (if 0, any serial is ok) */
+	unsigned int cur_serial; /* current serial in prox */
+	int cur_device_id;	/* current device ID in prox */
 	int maxWidth;		/* max active screen width in screen coords */
 	int maxHeight;		/* max active screen height in screen coords */
 	int leftPadding;	/* left padding for virtual tablet in device coordinates*/
@@ -239,17 +239,21 @@ struct _WacomDeviceRec
 	int reldn;
 	int wheelup;
 	int wheeldn;
+	int wheel2up;
+	int wheel2dn;
 	/* keystrokes assigned to wheel events (default is the buttons above).
-	 * Order is relup, reldwn, wheelup, wheeldn. */
-	unsigned wheel_keys[4][256];
+	 * Order is relup, reldwn, wheelup, wheeldn, wheel2up, wheel2dn.
+	 * Like 'keys', this array is one-indexed */
+	unsigned wheel_keys[6+1][256];
 
 	int striplup;
 	int stripldn;
 	int striprup;
 	int striprdn;
 	/* keystrokes assigned to strip events (default is the buttons above).
-	 * Order is striplup, stripldn, striprup, striprdn. */
-	unsigned strip_keys[4][256];
+	 * Order is striplup, stripldn, striprup, striprdn. Like 'keys', this
+	 * array is one-indexed */
+	unsigned strip_keys[4+1][256];
 	int nbuttons;           /* number of buttons for this subdevice */
 	int naxes;              /* number of axes */
 				/* FIXME: always 6, and the code relies on that... */
@@ -264,19 +268,19 @@ struct _WacomDeviceRec
 	int oldX;               /* previous X position */
 	int oldY;               /* previous Y position */
 	int oldZ;               /* previous pressure */
-	int oldCapacity;        /* previous capacity */
 	int oldTiltX;           /* previous tilt in x direction */
 	int oldTiltY;           /* previous tilt in y direction */    
 	int oldWheel;           /* previous wheel value */    
+	int oldWheel2;          /* previous wheel2 value */
 	int oldRot;             /* previous rotation value */
 	int oldStripX;          /* previous left strip value */
 	int oldStripY;          /* previous right strip value */
 	int oldThrottle;        /* previous throttle value */
 	int oldButtons;         /* previous buttons state */
 	int oldProximity;       /* previous proximity */
-	int oldHwProx;		/* previous hardware proximity */
+	int oldCursorHwProx;	/* previous cursor hardware proximity */
 	int old_device_id;	/* last in prox device id */
-	int old_serial;		/* last in prox tool serial number */
+	unsigned int old_serial;/* last in prox tool serial number */
 	int devReverseCount;	/* Relative ReverseConvert called twice each movement*/
 
 	/* JEJ - throttle */
@@ -290,14 +294,15 @@ struct _WacomDeviceRec
 	int minPressure;	/* the minimum pressure a pen may hold */
 
 	WacomToolPtr tool;         /* The common tool-structure for this device */
-	WacomToolAreaPtr toolarea; /* The area defined for this device */
 
 	int isParent;		/* set to 1 if the device is not auto-hotplugged */
 
 	/* property handlers to listen to for action properties */
 	Atom btn_actions[WCM_MAX_BUTTONS];
-	Atom wheel_actions[4];
+	Atom wheel_actions[6];
 	Atom strip_actions[4];
+
+	OsTimerPtr serial_timer; /* timer used for serial number property update */
 };
 
 /******************************************************************************
@@ -317,17 +322,16 @@ struct _WacomDeviceState
 	int y;
 	int buttons;
 	int pressure;
-	int capacity;
 	int tiltx;
 	int tilty;
 	int stripx;
 	int stripy;
 	int rotation;
 	int abswheel;
+	int abswheel2;
 	int relwheel;
 	int distance;
 	int throttle;
-	int discard_first;
 	int proximity;
 	int sample;	/* wraps every 24 days */
 };
@@ -384,7 +388,6 @@ extern WacomDeviceClass gWacomISDV4Device;
 
 #define TILT_REQUEST_FLAG       1
 #define TILT_ENABLED_FLAG       2
-#define RAW_FILTERING_FLAG      4
 
 #define MAX_CHANNELS 3
 #define PAD_CHANNEL (MAX_CHANNELS-1)
@@ -392,13 +395,11 @@ extern WacomDeviceClass gWacomISDV4Device;
 
 typedef struct {
 	int wcmZoomDistance;	       /* minimum distance for a zoom touch gesture */
-	int wcmZoomDistanceDefault;    /* default minimum distance for a zoom touch gesture */
 	int wcmScrollDistance;	       /* minimum motion before sending a scroll gesture */
 	int wcmScrollDirection;	       /* store the vertical or horizontal bit in use */
-	int wcmScrollDistanceDefault;  /* default minimum motion before sending a scroll gesture */
+	int wcmMaxScrollFingerSpread; /* maximum distance between fingers for scroll gesture */
 	int wcmGestureUsed;	       /* retain used gesture count within one in-prox event */
 	int wcmTapTime;	   	       /* minimum time between taps for a right click */
-	int wcmTapTimeDefault;         /* default minimum time between taps for a right click */
 } WacomGesturesParameters;
 
 enum WacomProtocol {
@@ -414,11 +415,15 @@ struct _WacomCommonRec
 	dev_t min_maj;               /* minor/major number */
 	unsigned char wcmFlags;     /* various flags (handle tilt) */
 	int debugLevel;
+	int vendor_id;		     /* Vendor ID */
 	int tablet_id;		     /* USB tablet ID */
-	int tablet_type;	     /* type (penabled/1FGT/2FGT, etc) of the tablet */
+	int tablet_type;	     /* bitmask of tablet features (WCM_LCD, WCM_PEN, etc) */
 	int fd;                      /* file descriptor to tablet */
 	int fd_refs;                 /* number of references to fd; if =0, fd is invalid */
 	unsigned long wcmKeys[NBITS(KEY_MAX)]; /* supported tool types for the device */
+	WacomDevicePtr wcmTouchDevice; /* The pointer for pen to access the
+					  touch tool of the same device id */
+	Bool wcmPenInProx;      /* Keep pen in-prox state for touch tool */
 
 	/* These values are in tablet coordinates */
 	int wcmMaxX;                 /* tablet max X value */
@@ -432,7 +437,6 @@ struct _WacomCommonRec
 	int wcmTouchResolY;	     /* touch Y resolution in points/m */
 	                             /* tablet Z resolution is equivalent
 	                              * to wcmMaxZ which is equal to 100% pressure */
-	int wcmMaxCapacity;	     /* max capacity value */
 	int wcmMaxDist;              /* tablet max distance value */
 	int wcmMaxtiltX;	     /* styli max tilt in X directory */ 
 	int wcmMaxtiltY;	     /* styli max tilt in Y directory */ 
@@ -440,10 +444,6 @@ struct _WacomCommonRec
 	int wcmMaxStripX;            /* Maximum fingerstrip X */
 	int wcmMaxStripY;            /* Maximum fingerstrip Y */
 
-	int nbuttons;                /* total number of buttons */
-	int npadkeys;                /* number of pad keys in the above array */
-	int padkey_code[WCM_MAX_BUTTONS];/* hardware codes for buttons */
-
 	WacomDevicePtr wcmDevices;   /* list of devices sharing same port */
 	int wcmPktLength;            /* length of a packet */
 	int wcmProtocolLevel;        /* Wacom Protocol used */
@@ -460,11 +460,7 @@ struct _WacomCommonRec
 	int wcmGesture;	     	     /* disable/enable touch gesture */
 	int wcmGestureDefault;       /* default touch gesture to disable when not supported */
 	int wcmGestureMode;	       /* data is in Gesture Mode? */
-	int wcmTouchpadMode;           /* in touchpad mode? */
 	WacomDeviceState wcmGestureState[MAX_FINGERS]; /* inital state when in gesture mode */
-	int wcmCapacity;	     /* disable/enable capacity */
-	int wcmCapacityDefault;      /* default to -1 when capacity isn't supported/disabled */
-				     /* 3 when capacity is supported */
 	WacomGesturesParameters wcmGestureParameters;
 	int wcmMaxCursorDist;	     /* Max mouse distance reported so far */
 	int wcmCursorProxoutDist;    /* Max mouse distance for proxy-out max/256 units */
@@ -478,13 +474,13 @@ struct _WacomCommonRec
 	void *private;		     /* backend-specific information */
 
 	WacomToolPtr wcmTool; /* List of unique tools */
+	WacomToolPtr serials; /* Serial numbers provided at startup*/
 
 	/* DO NOT TOUCH THIS. use wcmRefCommon() instead */
 	int refcnt;			/* number of devices sharing this struct */
 };
 
 #define HANDLE_TILT(comm) ((comm)->wcmFlags & TILT_ENABLED_FLAG)
-#define RAW_FILTERING(comm) ((comm)->wcmFlags & RAW_FILTERING_FLAG)
 
 /******************************************************************************
  * WacomTool
@@ -494,25 +490,11 @@ struct _WacomTool
 	WacomToolPtr next; /* Next tool in list */
 
 	int typeid; /* Tool type */
-	int serial; /* Serial id, 0 == no serial id */
+	unsigned int serial; /* Serial id, 0 == no serial id */
+	Bool enabled;
+	char *name;
 
-	WacomToolAreaPtr current;  /* Current area in-prox */
-	WacomToolAreaPtr arealist; /* List of defined areas */
-};
-
-/******************************************************************************
- * WacomToolArea
- *****************************************************************************/
-struct _WacomToolArea
-{
-	WacomToolAreaPtr next;
-
-	int topX;    /* Top X/Y */
-	int topY;
-	int bottomX; /* Bottom X/Y */
-	int bottomY;
-
-	InputInfoPtr device; /* The InputDevice connected to this area */
+	InputInfoPtr device; /* The InputDevice connected to this tool */
 };
 
 #endif /*__XF86_XF86WACOMDEFS_H */
